/*********************************************************************/
/* neblo_macros.h                                                   */
/* macros and small inline functions                                 */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#ifndef NEBLO_MACROS_H
#define NEBLO_MACROS_H

#include <neblo_.h>         /* for the data type defs */
#include <neblo_secret.h>  /* cipher stuff and defs */

/* gcc-specific macros */

#if __GNUC__ == 2 && __GNUC_MINOR__ < 96
#define __builtin_expect(x, expected_value) (x)
#endif

#ifndef likely
#define likely(x) __builtin_expect((x),1)
#endif
#ifndef unlikely
#define unlikely(x) __builtin_expect((x),0)
#endif


/* metric and topology in a 64-bit circular id. space */

/* distance in the circular 64-bit space */
#define _distance_(a,b) ((b)-(a))

/* does x belong to [a,b] in the circular 64-bit space? */
#define _belong_(x,a,b) \
( (a)<=(b) ? (a)<=(x) && (x)<=(b) : (b)>=(x) || (x)>=(a) )


/* safe invocation of various routines */

#define _do_(a,b,c) \
do { a; if (unlikely(b)) {perror(c);exit(1);} } while (0)


#ifdef USE_TALLOC

/* 
Use of thread-safe malloc 
by Paolo Gianrossi <gianrossi@disi.unige.it>
(2006-11-15)
*/
/* If we required talloc, simply use the allocation functions*/

#define S(a) do { a; } while (0)
#define _malloc_(addr, n) do{(addr)=xmalloc(n);}while(0)
#define _calloc_(addr, n) do{(addr)=xcalloc(1, n);} while(0)
#define _realloc_(newaddr, addr, n) do {(newaddr)=xrealloc(addr, n);} while(0)
#define _free_(addr) free(addr)

#else

#define S(a) \
do {\
  pthread_mutex_lock(&malloc_lock);\
  a;\
  pthread_mutex_unlock(&malloc_lock);\
} while (0)
#define _malloc_(addr,n) S(addr = malloc(n))
#define _calloc_(addr,n) S(addr = calloc(1,n))
#define _realloc_(newaddr,addr,n) S(newaddr = realloc(addr,n))
#define _free_(addr) S( if (addr) { free(addr); addr = NULL; } )

#endif /*USE_TALLOC*/


/* pthread init for mutexes and conds */

#define _pthmutexinit_(mutx) pthread_mutex_init(mutx,NULL)

#define _pthcondinit_(sem) pthread_cond_init(sem,NULL)


/* Miscellaneous */

#define min(a,b) ((a)<(b) ? (a):(b))
#define max(a,b) ((a)<(b) ? (b):(a))

#define _discardline_(line,fd)\
do {\
  if (line[strlen(line)-1] != '\n')\
    fgets(line,LINESIZE,fd);\
} while (0)

#define _read_(csock,buf,maxlen,rc,timeout,cont)\
{\
  __read__(csock,buf,maxlen,&rc,timeout);\
  if (unlikely(rc<0)) {cont;}\
}

#define _write_(csock,buf,len,rc,cont)\
{\
  __write__(csock,buf,len,&rc);\
  if (unlikely(rc<0)) {cont;}\
}

#define forall_peers(x,statement)\
for (;;) {\
  peer_t *x;\
  unsigned int i;\
  pthread_mutex_lock(&ring->lock);\
  for (x=ring->neigh_list; x!=NULL; x=x->next)\
    {statement;}\
  if (x == NULL) {\
    for (i=0; i<NUMFINGERS; i++) {\
      x = &ring->finger_table[i];\
          {statement;}\
    }\
  }\
  pthread_mutex_unlock(&ring->lock);\
  break;\
}

#define forall_neighbours(x,statement)\
for (;;) {\
  peer_t *x;\
  pthread_mutex_lock(&ring->lock);\
  for (x=ring->neigh_list; x!=NULL; x=x->next)\
    {statement;}\
  pthread_mutex_unlock(&ring->lock);\
  break;\
}

#define forall_neighbours_nolock(x,statement)\
for (;;) {\
  peer_t *x;\
  for (x=ring->neigh_list; x!=NULL; x=x->next)\
    {statement;}\
  break;\
}

#define forall_friends(x,statement)\
for (;;) {\
  peer_t *x;\
  pthread_mutex_lock(&ring->friend_lock);\
  for (x=ring->friend_list; x!=NULL; x=x->next)\
    {statement;}\
  pthread_mutex_unlock(&ring->friend_lock);\
  break;\
}

#define forall_friends_nolock(x,statement)\
for (;;) {\
  peer_t *x;\
  for (x=ring->friend_list; x!=NULL; x=x->next)\
    {statement;}\
  break;\
}

#define forall_oldfingers(x,statement)\
for (;;) {\
  peer_t *x;\
  pthread_mutex_lock(&ring->oldfinger_lock);\
  for (x=ring->oldfinger_list; x!=NULL; x=x->next)\
    {statement;}\
  pthread_mutex_unlock(&ring->oldfinger_lock);\
  break;\
}


#endif
